"""
An adapter for data generated on-the-fly.
"""

import itertools
import string
import urllib.parse
from collections.abc import Iterator
from datetime import datetime, timedelta, timezone
from typing import Any, Optional

from shillelagh.adapters.base import Adapter
from shillelagh.fields import Boolean, DateTime, Field, Float, Integer, Order, String
from shillelagh.filters import Filter
from shillelagh.typing import RequestedOrder, Row


def int_to_base26(n: int) -> str:
    """
    Convert 0→a, 1→b, …, 25→z, 26→aa, etc.
    """
    s = ""
    n += 1
    while n > 0:
        n, rem = divmod(n - 1, 26)
        s = string.ascii_lowercase[rem] + s
    return s


def make_seq(delta: Any) -> Iterator:
    """
    Generate a sequence of values starting from 0, incrementing by delta.
    """
    return (i * delta for i in itertools.count())  # pragma: no cover


DEFAULT_NUMBER_OF_ROWS = 10
TIME_RESOLUTIONS = {
    "microsecond": timedelta(microseconds=1),
    "millisecond": timedelta(milliseconds=1),
    "second": timedelta(seconds=1),
    "minute": timedelta(minutes=1),
    "hour": timedelta(hours=1),
    "day": timedelta(days=1),
    "week": timedelta(weeks=1),
    "month": timedelta(days=30),
    "quarter": timedelta(days=91),
    "year": timedelta(days=365),
    "decade": timedelta(days=3650),
    "century": timedelta(days=36500),
    "millennium": timedelta(days=365000),
}
TYPES = {  # pragma: no cover
    "int": itertools.count(),
    "float": make_seq(0.1),
    "str": (int_to_base26(i) for i in itertools.count()),
    "bool": (itertools.cycle([True, False])),
}
for type_, timedelta_ in TIME_RESOLUTIONS.items():
    TYPES[type_] = make_seq(timedelta_)


class VirtualMemory(Adapter):
    """
    An adapter for data generated on-the-fly:

        🍀> SELECT * FROM "virtual://?cols=id:int,t:day&rows=10";

    """

    safe = True

    supports_limit = False
    supports_offset = False
    supports_requested_columns = False

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = urllib.parse.urlparse(uri)
        return parsed.scheme == "virtual"

    @staticmethod
    def parse_uri(uri: str) -> tuple[dict[str, str], str, int]:
        parsed = urllib.parse.urlparse(uri)
        query_string = urllib.parse.parse_qs(parsed.query)

        cols = dict(
            col.split(":", 1)
            for col in (
                query_string["cols"][0].split(",") if "cols" in query_string else []
            )
        )
        cols = {k: v for k, v in cols.items() if v in TYPES}
        time_cols = {k: v for k, v in cols.items() if v in TIME_RESOLUTIONS}

        # default start to today, truncated to resolution
        if "start" in query_string:
            start = datetime.fromisoformat(query_string["start"][0])
        else:
            start = datetime.now(tz=timezone.utc)
            truncated = start.timestamp()
            for v in time_cols.values():
                truncated -= truncated % TIME_RESOLUTIONS[v].total_seconds()
            start = datetime.fromtimestamp(truncated, tz=timezone.utc)

        # if end is set, try to compute the number of rows
        if "end" in query_string and time_cols:
            end = datetime.fromisoformat(query_string["end"][0])
            resolution = max(TIME_RESOLUTIONS[v] for v in time_cols.values())
            rows = (end - start) // resolution

        # otherwise use the passed number of rows or default value
        else:
            rows = (
                int(query_string["rows"][0])
                if "rows" in query_string
                else DEFAULT_NUMBER_OF_ROWS
            )

        return cols, start.isoformat(), rows

    def __init__(
        self,
        cols: dict[str, str],
        start: str,
        rows: int,
    ):
        super().__init__()

        self.cols = cols
        self.start = datetime.fromisoformat(start)
        self.rows = rows

        self._set_columns()

    def _set_columns(self) -> None:
        type_map: dict[str, Field] = {
            "int": Integer(filters=None, order=Order.ASCENDING, exact=False),
            "float": Float(filters=None, order=Order.ASCENDING, exact=False),
            "str": String(filters=None, order=Order.ASCENDING, exact=False),
            "bool": Boolean(filters=None, order=Order.ASCENDING, exact=False),
        }
        self.columns = {
            k: type_map.get(
                v,
                DateTime(filters=None, order=Order.ASCENDING, exact=False),
            )
            for k, v in self.cols.items()
        }

    def get_columns(self) -> dict[str, Field]:
        return self.columns

    def get_data(
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        **kwargs: Any,
    ) -> Iterator[Row]:
        for row_id in range(self.rows):
            # pylint: disable=stop-iteration-return
            values = {v: next(TYPES[v]) for v in set(self.cols.values())}
            row = {"rowid": row_id}
            for k, v in self.cols.items():
                row[k] = self.start + values[v] if v in TIME_RESOLUTIONS else values[v]
            yield row



"""
A simple example showing the GSheets adapter.
"""

from shillelagh.backends.apsw.db import connect

if __name__ == "__main__":
    connection = connect(":memory:")
    cursor = connection.cursor()

    SQL = """
    SELECT *
    FROM "https://docs.google.com/spreadsheets/d/1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit#gid=1648320094"
    """
    for row in cursor.execute(SQL):
        print(row)



"""
A simple example querying a virtual table.
"""

from shillelagh.backends.apsw.db import connect

if __name__ == "__main__":
    connection = connect(":memory:")
    cursor = connection.cursor()

    SQL = """
        SELECT * FROM
        "virtual://?cols=a:int,b:str,c:bool,t1:day,t2:second&start=2024-01-01&end=2024-02-15";
    """
    for row in cursor.execute(SQL):
        print(row)



"""
A simple example showing the generic XML.
"""

import sys

from shillelagh.backends.apsw.db import connect

if __name__ == "__main__":
    API_KEY = sys.argv[1]

    connection = connect(":memory:")
    cursor = connection.cursor()

    SQL = f"""
    SELECT congress, type, latestAction FROM
    "https://api.congress.gov/v3/bill/118?format=xml&offset=0&limit=2&api_key={API_KEY}#.//bill"
    """
    for row in cursor.execute(SQL):
        print(row)



"""
An adapter for querying dbt Metric Flow metrics.

This is meant to be used with Apache Superset and has only been tested with the SQL that
it produces, eg:

    SELECT
        order_id__ordered_at__month AS order_id__ordered_at,
        order_id__is_food_order AS order_id__is_food_order,
        orders as orders
    FROM "https://semantic-layer.cloud.getdbt.com/"
    WHERE
        order_id__ordered_at >= '2017-01-01T00:00:00.000000'
        AND order_id__ordered_at < '2018-01-01T00:00:00.000000'
        AND order_id__is_food_order = true
    GROUP BY
        order_id__ordered_at__month,
        order_id__is_food_order
    ORDER BY order_id__ordered_at DESC
    LIMIT 10000

"""

import base64
import datetime
import inspect
import logging
import re
import time
from collections import defaultdict
from collections.abc import Iterator
from functools import lru_cache
from typing import Any, Optional, TypedDict, cast
from urllib.parse import urlparse

import pandas as pd
import pyarrow as pa
import sqlglot
from python_graphql_client import GraphqlClient
from sqlglot import expressions as exp
from sqlglot.optimizer.qualify_columns import qualify_columns
from sqlglot.optimizer.scope import traverse_scope

from shillelagh.adapters.base import Adapter
from shillelagh.backends.apsw.db import Cursor
from shillelagh.exceptions import ImpossibleFilterError, InternalError, ProgrammingError
from shillelagh.fields import (
    Boolean,
    Date,
    Decimal,
    Field,
    Filter,
    Integer,
    Order,
    String,
    Time,
    Unknown,
)
from shillelagh.filters import Equal, Impossible, IsNotNull, IsNull, NotEqual, Range
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

CUSTOM_URL_PATTERN = re.compile(
    r"""
        (?P<id>
            [a-zA-Z0-9]
            (?:
                [a-zA-Z0-9-]{0,61}
                [a-zA-Z0-9]
            )?
        )
        \.
        (?P<region>
            [a-zA-Z0-9]
            (?:
                [a-zA-Z0-9-]{0,61}
                [a-zA-Z0-9]
            )?
        )
        \.dbt.com
        $
    """,
    re.VERBOSE,
)

LIST_METRICS = """
    query GetMetrics($environmentId: BigInt!) {
        metrics(environmentId: $environmentId) {
            name
            description
            type
            dimensions {
                name
                description
                queryableGranularities
                type
                expr
            }
        }
    }
"""

CREATE_QUERY = """
    mutation GetData(
        $environmentId: BigInt!,
        $metrics: [MetricInput!]!,
        $where: [WhereInput!]!,
        $groupBy: [GroupByInput!]!,
        $orderBy: [OrderByInput!]!,
        $limit: Int,
    ) {
        createQuery(
            environmentId: $environmentId,
            metrics: $metrics,
            where: $where,
            groupBy: $groupBy,
            orderBy: $orderBy,
            limit: $limit,
        ) {
            queryId
        }
    }
"""

POLL_RESULTS = """
    query PollResults(
        $environmentId: BigInt!,
        $queryId: String!,
    ) {
        query(
            environmentId: $environmentId,
            queryId: $queryId,
        ) {
            sql
            status
            error
            arrowResult
        }
    }
"""

METRICS_FOR_DIMENSIONS = """
    query GetDimensions(
        $environmentId: BigInt!,
        $dimensions: [GroupByInput!]!,
    ) {
        metricsForDimensions(
            environmentId: $environmentId,
            dimensions: $dimensions,
        ) {
            name
        }
    }
"""

DIMENSIONS_FOR_METRICS = """
    query GetMetrics(
        $environmentId: BigInt!,
        $metrics: [MetricInput!]!,
    ) {
        dimensions(
            environmentId: $environmentId,
            metrics: $metrics,
        ) {
            name
        }
    }
"""


def find_cursor() -> Optional[Cursor]:
    """
    Inspects the stack up to find the cursor that called the adapter.

    This is ugly. It makes me want to quit tech and become a gardener.
    """
    for frame in inspect.stack():
        for value in frame.frame.f_locals.values():
            if isinstance(value, Cursor):
                return value

    return None


def extract_columns_from_sql(table: str, columns: set[str], sql: str) -> set[str]:
    """
    Parse the SQL and extract requested columns.
    """
    schema = {table: {column: "unused" for column in columns}}
    ast = sqlglot.parse_one(sql)
    ast = qualify_columns(ast, schema=schema)
    return {
        column.name
        for scope in traverse_scope(ast)
        for column in scope.columns
        if isinstance(scope.sources.get(column.table), exp.Table)
        and column.table == table
    }


class Timestamp(Field[pd.Timestamp, datetime.datetime]):
    """
    Pandas Timestamp.
    """

    type = "TIMESTAMP"
    db_api_type = "DATETIME"

    def parse(self, value: Optional[pd.Timestamp]) -> Optional[datetime.datetime]:
        return value.to_pydatetime() if value is not None else None

    def format(self, value: Optional[datetime.datetime]) -> Optional[pd.Timestamp]:
        return pd.Timestamp(value) if value is not None else None

    def quote(self, value: Optional[pd.Timestamp]) -> str:
        if value is None:
            return "NULL"
        return f"'{value.to_pydatetime().isoformat()}'"


class WhereInput(TypedDict):
    """
    Where parameter in the GraphQL API.
    """

    sql: str


class GroupByInput(TypedDict, total=False):
    """
    Group by parameter in the GraphQL API.
    """

    name: str
    grain: Optional[str]  # DAY, WEEK, MONTH, QUARTER, YEAR
    datePart: Optional[str]  # YEAR, QUARTER, MONTH, DAY, DOW, DOY


class MetricInput(TypedDict):
    """
    Metric input parameter in the GraphQL API.
    """

    name: str


class OrderByInput(TypedDict, total=False):
    """
    Order by parameter in the GraphQL API.
    """

    metric: MetricInput
    groupBy: GroupByInput
    descending: bool


def stream_to_dataframe(byte_string: str) -> pd.DataFrame:
    """
    Convert an Arrow stream to a Pandas DataFrame.
    """
    with pa.ipc.open_stream(base64.b64decode(byte_string)) as reader:
        return pa.Table.from_batches(reader, reader.schema).to_pandas()


class DbtMetricFlowAPI(Adapter):
    """
    An adapter for querying dbt Metric Flow metrics.
    """

    safe = True
    supports_limit = True
    supports_offset = False

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        """
        Is this a semantic layer URL?

        The adapter supports both the old single URL or custom user URLs.
        """
        parsed = urlparse(uri)
        return uri == "https://semantic-layer.cloud.getdbt.com/" or (
            parsed.scheme == "https" and bool(CUSTOM_URL_PATTERN.match(parsed.netloc))
        )

    @staticmethod
    def parse_uri(uri: str) -> tuple[str]:
        return (uri,)

    def __init__(self, table: str, service_token: str, environment_id: int):
        super().__init__()

        self.table = table
        self.environment_id = environment_id

        endpoint = self._get_endpoint(table)
        self.client = GraphqlClient(
            endpoint=endpoint,
            headers={"Authorization": f"Bearer {service_token}"},
        )

        self._set_columns()

    @staticmethod
    def _get_endpoint(url: str) -> str:
        """
        Return the GraphQL endpoint.
        """
        parsed = urlparse(url)
        if match := CUSTOM_URL_PATTERN.match(parsed.netloc):
            return f"https://{match['id']}.semantic-layer.{match['region']}.dbt.com/api/graphql"

        return "https://semantic-layer.cloud.getdbt.com/api/graphql"

    @lru_cache(maxsize=1)
    def _set_columns(self) -> None:
        payload = self.client.execute(
            query=LIST_METRICS,
            variables={"environmentId": self.environment_id},
        )

        self.columns: dict[str, Field] = {}
        self.metrics: dict[str, str] = {}
        self.dimensions: dict[str, str] = {}
        self.grains: dict[str, tuple[str, str]] = {}

        built_dimensions: set[str] = set()
        for metric in payload["data"]["metrics"]:
            self.columns[metric["name"]] = Decimal()
            self.metrics[metric["name"]] = metric["description"]

            for dimension in metric["dimensions"]:
                name = dimension["name"]
                if name in built_dimensions:
                    continue

                column = self._build_column_from_dimension(name)
                built_dimensions.add(name)

                # for time dimensions we create a dimension for each grain
                if dimension["type"] == "TIME":
                    for grain in dimension["queryableGranularities"]:
                        alias = f"{name}__{grain.lower()}"
                        self.columns[alias] = column
                        self.dimensions[alias] = dimension["description"]
                        self.grains[alias] = (name, grain.lower())
                else:
                    self.columns[name] = column
                    self.dimensions[name] = dimension["description"]

        self.columns = dict(sorted(self.columns.items()))

    def _run_query(self, **variables: Any) -> str:
        """
        Run a query, wait for it, and return Arrow payload.
        """
        payload = self.client.execute(query=CREATE_QUERY, variables=variables)
        if errors := payload.get("errors"):
            raise InternalError("\n\n".join(error["message"] for error in errors))

        query_id = payload["data"]["createQuery"]["queryId"]

        while True:
            payload = self.client.execute(
                query=POLL_RESULTS,
                variables={
                    "environmentId": self.environment_id,
                    "queryId": query_id,
                },
            )
            if error := payload["data"]["query"].get("error"):
                _logger.debug(payload["data"]["query"]["sql"])
                raise ProgrammingError(error)

            if payload["data"]["query"]["status"] == "SUCCESSFUL":
                break

            time.sleep(1)

        return cast(str, payload["data"]["query"]["arrowResult"])

    def _get_metrics_for_dimensions(self, dimensions: set[str]) -> set[str]:
        """
        Get metrics for a set of dimensions.
        """
        payload = self.client.execute(
            query=METRICS_FOR_DIMENSIONS,
            variables={
                "environmentId": self.environment_id,
                "dimensions": [{"name": dimension} for dimension in dimensions],
            },
        )

        return {
            metric["name"] for metric in payload["data"].get("metricsForDimensions", [])
        }

    def _get_dimensions_for_metrics(self, metrics: set[str]) -> set[str]:
        """
        Get dimensions for a set of metrics.
        """
        payload = self.client.execute(
            query=DIMENSIONS_FOR_METRICS,
            variables={
                "environmentId": self.environment_id,
                "metrics": [{"name": metric} for metric in metrics],
            },
        )

        reverse_grain: dict[str, set[str]] = defaultdict(set)
        for alias, (
            name,
            grain,  # pylint: disable=unused-variable
        ) in self.grains.items():
            reverse_grain[name].add(alias)

        dimensions = {"metric_time"}
        for dimension in payload["data"].get("dimensions", []):
            dimensions.add(dimension["name"])

        return dimensions

    def _build_column_from_dimension(  # pylint: disable=too-many-return-statements
        self,
        name: str,
    ) -> Field:
        """
        Build a field from a dbt dimension.

        Unfortunately the API does not provide the type of the dimension, so we
        need to fetch each one and read the type from the Arrow response.
        """
        try:
            byte_string = self._run_query(
                environmentId=self.environment_id,
                metrics=[],
                where=[],
                groupBy=[{"name": name}],
                orderBy=[],
                limit=0,
            )
        except ProgrammingError:
            return Unknown(
                filters=[Equal, NotEqual, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )

        with pa.ipc.open_stream(base64.b64decode(byte_string)) as reader:
            field = reader.schema[0]

        if pa.types.is_boolean(field.type):
            return Boolean(
                filters=[Equal, NotEqual, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_integer(field.type):
            return Integer(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_floating(field.type):
            return Decimal(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_string(field.type):
            return String(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_timestamp(field.type):
            return Timestamp(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_date(field.type):
            return Date(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_time(field.type):
            return Time(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_decimal(field.type):
            return Decimal(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )

        return Unknown(
            filters=[Equal, NotEqual, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )

    def _build_where(
        self,
        columns: dict[str, Field],
        bounds: dict[str, Filter],
    ) -> list[WhereInput]:
        """
        Build a ``WhereInput`` list from the bounds to filter in GraphQL.
        """
        where: list[WhereInput] = []
        for column_name, filter_ in bounds.items():
            if isinstance(filter_, Impossible):
                raise ImpossibleFilterError()

            if column_name in self.grains:
                base_column_name, grain = self.grains[column_name]
                ref = f"{{{{ TimeDimension('{base_column_name}', '{grain}') }}}}"
            else:
                ref = f"{{{{ Dimension('{column_name}') }}}}"

            field = columns[column_name]
            if isinstance(filter_, Equal):
                sql = f"{ref} = {field.quote(filter_.value)}"
            elif isinstance(filter_, NotEqual):
                sql = f"{ref} != {field.quote(filter_.value)}"
            elif isinstance(filter_, IsNull):
                sql = f"{ref} IS NULL"
            elif isinstance(filter_, IsNotNull):
                sql = f"{ref} IS NOT NULL"
            elif isinstance(filter_, Range):
                conditions = []
                if filter_.start is not None:
                    operator_ = ">=" if filter_.include_start else ">"
                    conditions.append(f"{ref} {operator_} {field.quote(filter_.start)}")
                if filter_.end is not None:
                    operator_ = "<=" if filter_.include_end else "<"
                    conditions.append(f"{ref} {operator_} {field.quote(filter_.end)}")
                sql = " AND ".join(conditions)
            else:
                raise ValueError(f"Invalid filter: {filter_}")

            where.append({"sql": sql})

        return where

    def _build_groupbys(
        self,
        requested_columns: set[str],
    ) -> list[GroupByInput]:
        """
        Build group bys based on the requested columns and the SQL query.
        """
        return [
            {"name": column}
            for column in requested_columns
            if column in self.dimensions
        ]

    def _build_orderbys(
        self,
        order: list[tuple[str, RequestedOrder]],
        groupbys: list[GroupByInput],
    ) -> list[OrderByInput]:
        """
        Build order bys based on the requested columns and the SQL query.
        """
        groupbys_map = {groupby["name"]: groupby for groupby in groupbys}

        orderbys = []
        for column, requested_order in order:
            orderby: OrderByInput = {
                "descending": requested_order == Order.DESCENDING,
            }

            if column in self.metrics:
                orderby["metric"] = {"name": column}
            elif column in self.dimensions:
                orderby["groupBy"] = groupbys_map[column]
            else:
                raise ProgrammingError(f"Invalid order by column: {column}")

            orderbys.append(orderby)

        return orderbys

    def get_columns(self) -> dict[str, Field]:
        return self.columns

    def get_data(
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        limit: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        cursor = find_cursor()
        if cursor is None or cursor.operation is None:
            raise InternalError("Unable to get reference to cursor")

        # When the virtual table has more than 63 columns apsw will return any requested
        # columns in the 1-63 range, plus ALL columns from 64-, since the mask is a 64-bit
        # value. This doesn't work for dbt, since only related columns should be requested
        # together. To solve this we need to parse the SQL and extract the columns
        # ourselves.
        requested_columns = extract_columns_from_sql(
            self.table,
            set(self.columns),
            cursor.operation,
        )

        groupbys = self._build_groupbys(requested_columns)
        orderbys = self._build_orderbys(order, groupbys)

        byte_string = self._run_query(
            environmentId=self.environment_id,
            metrics=[
                {"name": column}
                for column in requested_columns
                if column in self.metrics
            ],
            where=self._build_where(self.columns, bounds),
            groupBy=groupbys,
            orderBy=orderbys,
            limit=limit,
        )

        # pylint: disable=invalid-name
        df = stream_to_dataframe(byte_string)
        df.rename_axis("rowid", inplace=True)

        yield from df.reset_index().to_dict(orient="records")



"""
Custom functions available to the SQL backend.
"""

import importlib
import json
import sys
import time
from datetime import datetime, timedelta
from typing import Any, Optional

import pip
from packaging.version import InvalidVersion, Version

import shillelagh
from shillelagh.adapters.base import Adapter
from shillelagh.fields import FastISODateTime
from shillelagh.lib import find_adapter

if sys.version_info < (3, 10):
    from importlib_metadata import distribution
else:
    from importlib.metadata import distribution

__all__ = ["upgrade", "sleep", "get_metadata", "version", "date_trunc"]


def upgrade(target_version: str) -> str:
    """
    Upgrade the library to a given version.
    """
    try:
        pip.main(["install", f"shillelagh=={Version(target_version)}"])
        importlib.reload(shillelagh)
    except InvalidVersion:
        return f"Invalid version: {target_version}"
    except Exception as ex:  # pylint: disable=broad-except
        return f"Upgrade failed: {ex}"

    return f"Upgrade to {target_version} successful."


def sleep(seconds: int) -> None:
    """
    Sleep for ``n`` seconds.

    This is useful for troubleshooting timeouts::

        sql> SELECT sleep(60);

    """
    time.sleep(seconds)


def get_metadata(
    adapter_kwargs: dict[str, dict[str, Any]],
    adapters: list[type[Adapter]],
    uri: str,
) -> str:
    """
    Return metadata about a given table.

    Returns the name of the adapter that supports the table, as well
    as any extra metadata provided by the adapter::

        sql> SELECT GET_METADATA("https://docs.google.com/spreadsheets/d/1/edit#gid=0");
        GET_METADATA("https://docs.google.com/spreadsheets/d/1/edit#gid=0")
        -------------------------------------------------------------------
        {
            "extra": {
                "Spreadsheet title": "A title",
                "Sheet title": "Another title"
            },
            "adapter": "GSheetsAPI"
        }

    """
    adapter, args, kwargs = find_adapter(uri, adapter_kwargs, adapters)
    instance = adapter(*args, **kwargs)

    return json.dumps(
        {
            "extra": instance.get_metadata(),
            "adapter": adapter.__name__,
        },
    )


def version() -> str:
    """
    Return the current version of Shillelagh.

    As an example::

        sql> SELECT VERSION();
        VERSION()
        -----------
        0.7.4

    """
    return str(distribution("shillelagh").version)


def date_trunc(  # pylint: disable=too-many-return-statements
    value: Optional[str],
    unit: str,
) -> Optional[str]:
    """
    Truncate a datetime to a given unit.
    """
    field = FastISODateTime()
    timestamp = field.parse(value)
    if timestamp is None:
        return None

    unit = unit.lower()
    if unit == "year":
        truncated = datetime(year=timestamp.year, month=1, day=1)
    elif unit == "quarter":
        month = ((timestamp.month - 1) // 3) * 3 + 1
        truncated = datetime(year=timestamp.year, month=month, day=1)
    elif unit == "month":
        truncated = datetime(year=timestamp.year, month=timestamp.month, day=1)
    elif unit == "week":
        # assumes the week starts on Monday
        start_of_week = timestamp - timedelta(days=timestamp.weekday())
        truncated = datetime(
            year=start_of_week.year,
            month=start_of_week.month,
            day=start_of_week.day,
        )
    elif unit == "day":
        truncated = datetime(
            year=timestamp.year,
            month=timestamp.month,
            day=timestamp.day,
        )
    elif unit == "hour":
        truncated = datetime(
            year=timestamp.year,
            month=timestamp.month,
            day=timestamp.day,
            hour=timestamp.hour,
        )
    elif unit == "minute":
        truncated = datetime(
            year=timestamp.year,
            month=timestamp.month,
            day=timestamp.day,
            hour=timestamp.hour,
            minute=timestamp.minute,
        )
    elif unit == "second":
        truncated = datetime(
            year=timestamp.year,
            month=timestamp.month,
            day=timestamp.day,
            hour=timestamp.hour,
            minute=timestamp.minute,
            second=timestamp.second,
        )
    else:
        raise ValueError(f"Unsupported truncation unit: {unit}")

    return field.format(truncated)



# pylint: disable=invalid-name, c-extension-no-member, unused-import
"""
A base DB API 2.0 implementation.
"""

import itertools
import re
from collections.abc import Iterator
from functools import wraps
from typing import Any, Callable, Generic, Optional, TypeVar, cast

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import (  # nopycln: import; pylint: disable=redefined-builtin
    DatabaseError,
    DataError,
    Error,
    IntegrityError,
    InterfaceError,
    InternalError,
    NotSupportedError,
    OperationalError,
    ProgrammingError,
    Warning,
)
from shillelagh.types import (
    BINARY,
    DATETIME,
    NUMBER,
    ROWID,
    STRING,
    Binary,
    Date,
    DateFromTicks,
    Time,
    TimeFromTicks,
    Timestamp,
    TimestampFromTicks,
)
from shillelagh.typing import Description

__all__ = [
    "DatabaseError",
    "DataError",
    "Error",
    "IntegrityError",
    "InterfaceError",
    "InternalError",
    "OperationalError",
    "BINARY",
    "DATETIME",
    "NUMBER",
    "ROWID",
    "STRING",
    "Binary",
    "Date",
    "DateFromTicks",
    "Time",
    "TimeFromTicks",
    "Timestamp",
    "TimestampFromTicks",
    "Warning",
    "apilevel",
    "threadsafety",
    "paramstyle",
]

apilevel = "2.0"
threadsafety = 2
paramstyle = "qmark"

NO_SUCH_TABLE = re.compile("no such table: (?P<uri>.*)")
DEFAULT_SCHEMA = "main"

CURSOR_METHOD = TypeVar("CURSOR_METHOD", bound=Callable[..., Any])


def check_closed(method: CURSOR_METHOD) -> CURSOR_METHOD:
    """Decorator that checks if a connection or cursor is closed."""

    @wraps(method)
    def wrapper(self: "Cursor", *args: Any, **kwargs: Any) -> Any:
        if self.closed:
            raise ProgrammingError(f"{self.__class__.__name__} already closed")
        return method(self, *args, **kwargs)

    return cast(CURSOR_METHOD, wrapper)


def check_result(method: CURSOR_METHOD) -> CURSOR_METHOD:
    """Decorator that checks if the cursor has results from ``execute``."""

    @wraps(method)
    def wrapper(self: "Cursor", *args: Any, **kwargs: Any) -> Any:
        if self._results is None:  # pylint: disable=protected-access
            raise ProgrammingError("Called before ``execute``")
        return method(self, *args, **kwargs)

    return cast(CURSOR_METHOD, wrapper)


class Cursor:  # pylint: disable=too-many-instance-attributes
    """
    Connection cursor.
    """

    def __init__(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        adapters: list[type[Adapter]],
        adapter_kwargs: dict[str, dict[str, Any]],
        schema: str = DEFAULT_SCHEMA,
    ):
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs
        self.schema = schema

        # This read/write attribute specifies the number of rows to fetch at a
        # time with .fetchmany(). It defaults to 1 meaning to fetch a single
        # row at a time.
        self.arraysize = 1

        self.closed = False

        # this is updated only after a query
        self.description: Description = None

        # this is set to an iterator of rows after a successful query
        self._results: Optional[Iterator[tuple[Any, ...]]] = None
        self._rowcount = -1

        self.operation: Optional[str] = None

    @property  # type: ignore
    @check_closed
    def rowcount(self) -> int:
        """
        Return the number of rows after a query.
        """
        try:
            results = list(self._results)  # type: ignore
        except TypeError:
            return -1

        n = len(results)
        self._results = iter(results)
        return max(0, self._rowcount) + n

    @check_closed
    def close(self) -> None:
        """
        Close the cursor.
        """
        self.closed = True

    @check_closed
    def execute(
        self,
        operation: str,
        parameters: Optional[tuple[Any, ...]] = None,
    ) -> "Cursor":
        """
        Execute a query using the cursor.
        """
        raise NotImplementedError()

    @check_closed
    def executemany(
        self,
        operation: str,
        seq_of_parameters: Optional[list[tuple[Any, ...]]] = None,
    ) -> "Cursor":
        """
        Execute multiple statements.

        Currently not supported.
        """
        raise NotSupportedError(
            "``executemany`` is not supported, use ``execute`` instead",
        )

    @check_result
    @check_closed
    def fetchone(self) -> Optional[tuple[Any, ...]]:
        """
        Fetch the next row of a query result set, returning a single sequence,
        or ``None`` when no more data is available.
        """
        try:
            row = self.next()
        except StopIteration:
            return None

        self._rowcount = max(0, self._rowcount) + 1

        return row

    @check_result
    @check_closed
    def fetchmany(self, size=None) -> list[tuple[Any, ...]]:
        """
        Fetch the next set of rows of a query result, returning a sequence of
        sequences (e.g. a list of tuples). An empty sequence is returned when
        no more rows are available.
        """
        size = size or self.arraysize
        results = list(itertools.islice(self, size))

        return results

    @check_result
    @check_closed
    def fetchall(self) -> list[tuple[Any, ...]]:
        """
        Fetch all (remaining) rows of a query result, returning them as a
        sequence of sequences (e.g. a list of tuples). Note that the cursor's
        arraysize attribute can affect the performance of this operation.
        """
        results = list(self)

        return results

    @check_closed
    def setinputsizes(self, sizes: int) -> None:
        """
        Used before ``execute`` to predefine memory areas for parameters.

        Currently not supported.
        """

    @check_closed
    def setoutputsizes(self, sizes: int) -> None:
        """
        Set a column buffer size for fetches of large columns.

        Currently not supported.
        """

    @check_result
    @check_closed
    def __iter__(self) -> Iterator[tuple[Any, ...]]:
        for row in self._results:  # type: ignore
            self._rowcount = max(0, self._rowcount) + 1
            yield row

    @check_result
    @check_closed
    def __next__(self) -> tuple[Any, ...]:
        return next(self._results)  # type: ignore

    next = __next__


ConnectionCursor = TypeVar("ConnectionCursor", bound=Cursor)


class Connection(
    Generic[ConnectionCursor],
):  # pylint: disable=too-many-instance-attributes
    """Connection."""

    def __init__(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        adapters: list[type[Adapter]],
        adapter_kwargs: dict[str, dict[str, Any]],
        schema: str = DEFAULT_SCHEMA,
        safe: bool = False,
    ):
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs
        self.schema = schema
        self.safe = safe

        self.closed = False
        self.cursors: list[ConnectionCursor] = []

    @check_closed
    def close(self) -> None:
        """Close the connection now."""
        self.closed = True
        for cursor in self.cursors:
            if not cursor.closed:
                cursor.close()

    @check_closed
    def commit(self) -> None:
        """Commit any pending transaction to the database."""

    @check_closed
    def rollback(self) -> None:
        """Rollback any transactions."""

    @check_closed
    def cursor(self) -> ConnectionCursor:
        """Return a new Cursor Object using the connection."""
        raise NotImplementedError(
            "Subclasses must implement the `cursor` method",
        )

    @check_closed
    def execute(
        self,
        operation: str,
        parameters: Optional[tuple[Any, ...]] = None,
    ) -> "Cursor":
        """
        Execute a query on a cursor.
        """
        cursor = self.cursor()
        return cursor.execute(operation, parameters)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.commit()
        self.close()



"""
Parse and format Google Sheet number formats.

https://developers.google.com/sheets/api/guides/formats#number_format_tokens
"""

# pylint: disable=c-extension-no-member, broad-exception-raised
import math
import operator
import re
from collections.abc import Iterator
from itertools import zip_longest
from typing import Any, Union, cast

from shillelagh.adapters.api.gsheets.parsing.base import (
    LITERAL,
    InvalidValue,
    Token,
    tokenize,
)


def adjust_value(value: Union[int, float], tokens: list[Token]) -> Union[int, float]:
    """
    Adjust value applying percent, scientific notation, and comma multiplier.
    """
    for token in tokens:
        if token.__class__.__name__ == "PERCENT":
            value *= 100
        elif token.__class__.__name__ == "COMMA":
            value /= 1000 ** len(token.token)
        elif token.__class__.__name__ == "E":
            while value > 10:
                value /= 10
            while value < 1:
                value *= 10

    return value


class DIGITS(Token):
    """
    One or more digits in the number.

    Each token renders digits differently:

    - 0: if the digit is an insignificant 0, it is rendered as 0.
    - #: if the digit is an insignificant 0, it is not rendered.
    - ?: if the digit is an insignificant 0, it is rendered as a space.

    If there's a comma between tokens it renders the entire number with grouping
    separators (grouping by the thousands)
    """

    regex = r"(0|#|\?)+(,(0|#|\?){3})?"

    def format(  # pylint: disable=too-many-branches
        self,
        value: Union[int, float],
        tokens: list[Token],
    ) -> str:
        value = adjust_value(value, tokens)
        number = str(value)

        formatted: list[str] = []
        if is_fractional(self, tokens):
            number = number.split(".")[1] if "." in number else "0"
            for token, digit in zip_longest(self.token, number):
                if token is None:
                    break
                if token == "0":
                    formatted.append(digit or "0")
                elif token == "#":
                    formatted.append(digit or "")
                elif token == "?":
                    formatted.append(digit or " ")
                else:
                    raise Exception(f"Invalid token: {token}")
            return "".join(formatted)

        number = number.split(".", maxsplit=1)[0]
        has_comma = "," in self.token
        token = self.token.replace(",", "")

        for i, (token, digit) in enumerate(zip_longest(token[::-1], number[::-1])):
            if token is None:
                formatted.append(digit)
            elif token == "0":
                formatted.append(digit or "0")
            elif token == "#":
                formatted.append(digit or "")
            elif token == "?":
                formatted.append(digit or " ")
            else:
                raise Exception(f"Invalid token: {token}")

            # group by thousands
            if has_comma and (i + 1) % 3 == 0:
                formatted.append(",")

        return "".join(formatted[::-1]).strip(",")

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        match = re.match(r"\s*(\+|-)?\d+(,\d+)*", value)
        if not match:
            raise InvalidValue(value)

        digits = match.group().replace(",", "")
        if is_fractional(self, tokens):
            return (
                {
                    "operation": lambda number: math.copysign(
                        abs(number) + get_fraction(digits),
                        number,
                    ),
                },
                value[len(match.group()) :],
            )

        return (
            {"operation": lambda number: number + int(digits)},
            value[len(match.group()) :],
        )


class PERIOD(Token):
    """
    The decimal point in the number.

    The first period represents the decimal point in the number. Subsequent
    periods are rendered as literals. If you include a decimal point in the
    format, it will always be rendered, even for whole numbers.
    """

    regex = r"\."

    def format(self, value: Union[int, float], tokens: list[Token]) -> str:
        return "."

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        match = re.match(r"\.", value)
        if not match:
            raise InvalidValue(value)

        return {}, value[1:]


class PERCENT(Token):
    """
    Literal percent sign.

    Appears as a literal but also causes existing numbers to be multiplied
    by 100 before being rendered, in order to make percentages more readable.
    """

    regex = "%"

    def format(self, value: Union[int, float], tokens: list[Token]) -> str:
        return "%"

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        match = re.match("%", value)
        if not match:
            raise InvalidValue(value)

        return {"operation": lambda number: number / 100.0}, value[1:]


class COMMA(Token):
    """
    Multiplier.

    When appearing after the digit characters the comma scales the digits by one thousand
    per comma.
    """

    regex = ",+"

    def format(self, value: Union[int, float], tokens: list[Token]) -> str:
        return ""

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        size = len(self.token)
        return {"operation": lambda number: number * 1000**size}, value


class E(Token):  # pylint: disable=invalid-name
    """
    Scientific format.

    Renders the number in scientific format, with the formatting to the left of the E
    used for the non-exponent portion and the formatting to the right of the E used for
    the exponent portion. E+ will show a + sign for positive exponents. E- will only
    show a sign for negative exponents. If lowercase is used, the output e is
    lowercased as well.
    """

    regex = r"(E|e)(-|\+)((0|#|\?)+)"

    def format(self, value: Union[int, float], tokens: list[Token]) -> str:
        exponent = 0

        if value >= 1:
            sign = "+" if "+" in self.token else ""
            while value > 10:
                value /= 10
                exponent += 1
        else:
            sign = "-"
            while value < 1:
                value *= 10
                exponent += 1

        match = re.match(self.regex, self.token)
        if not match:
            # should never happen
            raise Exception("You are likely to be eaten by a grue.")
        cased_e = match.group(1)

        # process the exponent according to the pattern
        pattern = match.group(3)
        formatted_exponent = format_number_pattern(exponent, pattern)

        return f"{cased_e}{sign}{formatted_exponent}"

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        match = re.match(r"(E|e)((\+|-)?\d+)", value)
        if not match:
            raise InvalidValue(value)

        exponent = int(match.group(2))
        size = len(match.group())

        return {"operation": lambda number: number * 10**exponent}, value[size:]


class FRACTION(Token):
    """
    Fractional format.

    If it appears between two digit characters (0, # or ?), then it treats those digit
    groups as a fractional format. For example, the number format 0 #/# renders the
    number 23.25 as 23 1/4. The denominator can also be a literal integer, in which case
    it will enforce that integer as the denominator. The number format 0 #/8 displays
    the number 23.25 as 23 2/8. The fraction part is not rendered at all if the
    numerator would become 0. The number 23.1 with the number format 0 #/3 renders as
    just 23 (because the 0.1 rounded to 0/3). / is not compatible with scientific format
    or a format with a decimal point in it.
    """

    regex = r"(?:((?:0|#|\?)+))/(?:((?:0|#|\?)+)|(\d+))"

    def format(self, value: Union[int, float], tokens: list[Token]) -> str:
        fractional_part = float(value - math.floor(value))
        numerator: Union[int, float]
        numerator, denominator = fractional_part.as_integer_ratio()

        # force denominator?
        match = re.match(self.regex, self.token)
        if not match:
            # should never happen
            raise Exception("You are likely to be eaten by a grue.")
        groups = match.groups()
        if groups[2] is not None:
            numerator *= int(groups[2]) / denominator
            formatted_denominator = groups[2]
        else:
            # the denominator needs to be formatted as a fraction, with spaces and zeros
            # right padded
            pattern = "." + groups[1]
            number = get_fraction(str(denominator))
            formatted_denominator = format_number_pattern(number, pattern).lstrip(".")

        if math.floor(numerator) == 0:
            return ""

        formatted_numerator = format_number_pattern(numerator, groups[0])

        return f"{formatted_numerator}/{formatted_denominator}"

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        match = re.match(r"\s*(\d+)/(\d+)", value)
        if not match:
            # fractions with numerator 0 are not shown
            return {}, value

        numerator = int(match.group(1))
        denominator = int(match.group(2))
        size = len(match.group())

        return {"operation": lambda number: number + (numerator / denominator)}, value[
            size:
        ]


class STAR(Token):
    """
    This is included for compatibility with Excel number formats. It is currently ignored.
    """

    regex = r"\*"

    def format(self, value: Union[int, float], tokens: list[Token]) -> str:
        return ""

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        return {}, value


class UNDERSCORE(Token):
    """
    Skips the next character and renders a space.

    This is used to line up number formats where the negative value is surrounded
    by parenthesis.
    """

    regex = "_."

    def format(self, value: Union[int, float], tokens: list[Token]) -> str:
        return " "

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        if value[:1] != " ":
            raise InvalidValue(value[:1])

        return {}, value[1:]


class AT(Token):
    """
    Inserts the raw text for the cell, if the cell has text input.

    Not compatible with any of the other special characters and won’t display for
    numeric values (which are displayed as general format).
    """

    regex = "@"

    def format(self, value: Union[int, float, str], tokens: list[Token]) -> str:
        return str(value)

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        i = tokens.index(self)

        if i == len(tokens) - 1:
            # last token, consume everything
            return {"operation": lambda number: value}, ""

        # LL(1) to see how much can be consumed
        sibling = tokens[i + 1]
        j = 0
        for j in range(len(value)):
            try:
                sibling.parse(value[j:], tokens)
                break
            except InvalidValue:
                continue
        else:
            raise InvalidValue(value)

        return {"operation": lambda number: value[:j]}, value[j:]


class COLOR(Token):
    """
    Causes any value that is rendered by this sub-format to appear with the given text color.

    Valid values for Color are Black, Blue, Cyan, Green, Magenta, Red, White or Yellow.
    Valid values for the "#" in Color# are 0 - 56 (this color palette shows a list of the
    colors that correspond to each number). Number format colors will override any
    user-entered colors on the cell, but will not override colors set by conditional
    formatting.
    """

    regex = r"\[(Black|Blue|Cyan|Green|Magenta|Red|White|Yellow|Color\d{1,2})\]"

    def format(self, value: Union[int, float], tokens: list[Token]) -> str:
        return ""

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        return {}, value


class CONDITION(Token):
    """
    Replaces the default comparison section with another conditional expression.

    Replaces the default positive, negative, or zero comparison section with another
    conditional expression. For example, [<100]”Low”;[>1000]”High”;000 will render the
    word “Low” for values below 100, “High” for values above 1000 and a three digit
    number (with leading 0s) for anything in between. Conditions can only be applied to
    the first two sub-formats and if a number matches more than one, it will use the
    first one it matches. If there is a third format, it will be used for "everything
    else", otherwise if a number doesn’t match either format, it will be rendered as all
    "#"s filling up the cell width. The fourth format is always used for text, if it
    exists.
    """

    regex = r"\[(>|>=|<|<=|=)\d*(\.\d*)?\]"

    def format(self, value: Union[int, float], tokens: list[Token]) -> str:
        return ""

    def parse(self, value: str, tokens: list[Token]) -> tuple[dict[str, Any], str]:
        return {}, value


def get_fraction(number: str) -> float:
    """
    Return the fraction associated with a fractional part.

        >>> get_fraction("9")
        0.9
        >>> get_fraction("123")
        0.123
        >>> get_fraction("1")
        0.1
        >>> get_fraction("001")
        0.001

    """
    if int(number) < 0:
        raise Exception("Number should be a positive integer")

    return cast(float, int(number) / (10 ** len(number)))


def parse_number_pattern(value: str, pattern: str) -> float:
    """
    Parse a value using a given pattern.
    """
    formats = pattern.split(";")

    i = -1
    for i, format_ in enumerate(formats):
        try:
            number = parse_number_format(value, format_)
            break
        except InvalidValue:
            # weird edge case
            if value.startswith("-"):
                try:
                    number = -parse_number_pattern(value[1:], format_)
                    break
                except InvalidValue:
                    pass
    else:
        try:
            return int(value)
        except ValueError:
            try:
                return float(value)
            except ValueError as ex:
                raise InvalidValue(
                    f"Unable to parse value {value} with pattern {pattern}",
                ) from ex

    # is negative?
    if i == 1 and (len(formats) > 2 or (len(formats) == 2 and "@" not in formats[1])):
        return -number

    return number


def fix_periods(tokens: Iterator[Token]) -> Iterator[Token]:
    """
    Convert periods into literals after the first one.
    """
    seen_period = False
    for token in tokens:
        if token.__class__.__name__ == "PERIOD":
            yield LITERAL(".") if seen_period else token
            seen_period = True
        else:
            yield token


def parse_number_format(value: str, format_: str) -> float:
    """
    Parse a value using a given format pattern.
    """
    classes = [
        FRACTION,  # should come first
        DIGITS,
        COMMA,
        PERIOD,
        PERCENT,
        E,
        STAR,
        UNDERSCORE,
        AT,
        COLOR,
        CONDITION,
        LITERAL,
    ]

    number = 0

    tokens = list(fix_periods(tokenize(format_, classes)))
    for token in tokens:
        consumed, value = token.parse(value, tokens)
        if "operation" in consumed:
            number = consumed["operation"](number)

    return number


def has_condition(pattern: str) -> bool:
    """
    Return true if the pattern has condition instructions.
    """
    return bool(re.match(r"\[(>|>=|<|<=|=)\d*(\.\d*)?\]", pattern))


def condition_matches(value: Union[int, float], format_: str) -> bool:
    """
    Return true if the value matches the condition in the pattern.
    """
    match = re.search(r"\[(>|>=|<|<=|=)(\d*(?:\.\d*)?)\]", format_)
    if not match:
        return True

    operators = {
        ">": operator.gt,
        ">=": operator.ge,
        "<": operator.lt,
        "<=": operator.le,
        "=": operator.eq,
    }
    comparison, threshold = match.groups()
    op = operators[comparison]  # pylint: disable=invalid-name
    return bool(op(float(value), float(threshold)))


def format_number_pattern(  # pylint: disable=too-many-branches
    value: Union[int, float, str],
    pattern: str,
) -> str:
    """
    Format a number to a given pattern.
    """
    formats = pattern.split(";")

    if not any(format_ for format_ in formats):
        raise Exception("Empty pattern!")

    if isinstance(value, str):
        for format_ in formats:
            if "@" in format_:
                break
        else:
            raise Exception("No text format found for string value")

    elif has_condition(pattern):
        for format_ in formats:
            if condition_matches(value, format_):
                break
        else:
            format_ = '"########"'

    else:
        # remove any text format
        formats = [format_ for format_ in formats if "@" not in format_]

        if value > 0:
            format_ = formats[0]
        elif value == 0:
            format_ = formats[2] if len(formats) == 3 else formats[0]
        else:
            if len(formats) >= 2:
                format_ = formats[1]
                value *= -1
            else:
                format_ = formats[0]

    classes = [
        FRACTION,  # should come first
        DIGITS,
        COMMA,
        PERIOD,
        PERCENT,
        E,
        STAR,
        UNDERSCORE,
        AT,
        COLOR,
        CONDITION,
        LITERAL,
    ]

    parts = []

    tokens = list(fix_periods(tokenize(format_, classes)))
    for token in tokens:
        parts.append(token.format(value, tokens))

    return "".join(parts)


def is_fractional(token: Token, tokens: list[Token]) -> bool:
    """
    Return true if the token is after the period.
    """
    seen_self = False
    for sibling in tokens:
        if sibling is token:
            seen_self = True
        elif isinstance(sibling, PERIOD) and not seen_self:
            return True
    return False



"""
An adapter for GitHub.
"""

import json
import logging
import urllib.parse
from collections.abc import Iterator
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Optional, TypedDict

import jsonpath

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Boolean, DateTime, Field, Integer, String, StringDateTime
from shillelagh.filters import Equal, Filter
from shillelagh.lib import get_session
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

PAGE_SIZE = 100


class JSONString(Field[Any, str]):
    """
    A field to handle JSON values.
    """

    type = "TEXT"
    db_api_type = "STRING"

    def parse(self, value: Any) -> Optional[str]:
        return value if value is None else json.dumps(value)


@dataclass
class Column:
    """
    A class to track columns in GitHub resources.
    """

    # The name of the column in the virtual table, eg, ``userid``.
    name: str

    # The JSON path to the value in the GitHub response. Eg, user ID is
    # returned as ``{"user": {"id": 42}}``, so the JSON path would be
    # ``user.id``.
    json_path: str

    # The Shillelagh field to be used for the column.
    field: Field

    # A default value for when the column is not specified. Eg, for ``pulls``
    # the API defaults to showing only PRs with an open state, so we need to
    # default the column to ``all`` to fetch all PRs when state is not
    # specified in the query.
    default: Optional[Filter] = None


def participation_processor(payload: dict[str, Any]) -> list[dict[str, Any]]:
    """
    Process participation data.

    https://docs.github.com/en/rest/metrics/statistics?apiVersion=2022-11-28#get-the-weekly-commit-count
    """
    today_utc_midnight = datetime.now(timezone.utc).replace(
        hour=0,
        minute=0,
        second=0,
        microsecond=0,
    )
    start = today_utc_midnight - timedelta(weeks=len(payload["all"]))

    return [
        {
            "start_at": start + timedelta(weeks=i),
            "end_at": start + timedelta(weeks=i + 1),
            "all": all,
            "owner": owner,
        }
        for i, (all, owner) in enumerate(zip(payload["all"], payload["owner"]))
    ]


class EndPointDefinition(TypedDict):
    """
    A definition for an endpoint.

    This is used to define the columns and the path to the values in the JSON response.
    It can also specify if the endpoint is paginated (most are) and a processor to
    transform the payload.
    """

    columns: list[Column]
    paginated: bool
    processor: Optional[Callable[[dict[str, Any]], list[dict[str, Any]]]]


# a mapping from the column name (eg, ``userid``) to the path in the JSON
# response (``{"user": {"id": 42}}`` => ``user.id``) together with the field
TABLES: dict[str, dict[str, EndPointDefinition]] = {
    "repos": {
        "pulls": {
            "columns": [
                Column("url", "html_url", String()),
                Column("id", "id", Integer()),
                Column("number", "number", Integer(filters=[Equal])),
                Column("state", "state", String(filters=[Equal]), Equal("all")),
                Column("title", "title", String()),
                Column("userid", "user.id", Integer()),
                Column("username", "user.login", String()),
                Column("draft", "draft", Boolean()),
                Column("head", "head.ref", String(filters=[Equal])),  # head.label?
                Column("created_at", "created_at", StringDateTime()),
                Column("updated_at", "updated_at", StringDateTime()),
                Column("closed_at", "closed_at", StringDateTime()),
                Column("merged_at", "merged_at", StringDateTime()),
            ],
            "paginated": True,
            "processor": None,
        },
        "issues": {
            "columns": [
                Column("url", "html_url", String()),
                Column("id", "id", Integer()),
                Column("number", "number", Integer(filters=[Equal])),
                Column("state", "state", String(filters=[Equal]), Equal("all")),
                Column("title", "title", String()),
                Column("userid", "user.id", Integer()),
                Column("username", "user.login", String()),
                Column("draft", "draft", Boolean()),
                Column("locked", "locked", Boolean()),
                Column("comments", "comments", Integer()),
                Column("created_at", "created_at", StringDateTime()),
                Column("updated_at", "updated_at", StringDateTime()),
                Column("closed_at", "closed_at", StringDateTime()),
                Column("body", "body", String()),
                Column("author_association", "author_association", String()),
                Column("labels", "labels[*].name", JSONString()),
                Column("assignees", "assignees[*].login", JSONString()),
                Column("reactions", "reactions", JSONString()),
            ],
            "paginated": True,
            "processor": None,
        },
        "stats/punch_card": {
            "columns": [
                Column("dow", "$.[0]", Integer()),
                Column("hour", "$.[1]", Integer()),
                Column("commits", "$.[2]", Integer()),
            ],
            "paginated": True,
            "processor": None,
        },
        "stats/participation": {
            "columns": [
                Column("start_at", "$.start_at", DateTime()),
                Column("end_at", "$.end_at", DateTime()),
                Column("all", "$.all", Integer()),
                Column("owner", "$.owner", Integer()),
            ],
            "paginated": False,
            "processor": participation_processor,
        },
    },
}


class GitHubAPI(Adapter):
    """
    An adapter for GitHub.
    """

    safe = True

    supports_limit = False
    supports_offset = False

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = urllib.parse.urlparse(uri)

        if parsed.path.count("/") < 4:
            return False

        # pylint: disable=unused-variable
        _, base, owner, repo, resource = parsed.path.split("/", 4)
        return (
            parsed.netloc == "api.github.com"
            and base in TABLES
            and resource in TABLES[base]
        )

    @staticmethod
    def parse_uri(uri: str) -> tuple[str, str, str, str]:
        parsed = urllib.parse.urlparse(uri)
        _, base, owner, repo, resource = parsed.path.split("/", 4)
        return (
            base,
            owner,
            repo,
            resource,
        )

    def __init__(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        base: str,
        owner: str,
        repo: str,
        resource: str,
        access_token: Optional[str] = None,
    ):
        super().__init__()

        self.base = base
        self.owner = owner
        self.repo = repo
        self.resource = resource
        self.access_token = access_token

        self._session = get_session(
            request_headers={},
            cache_name="github_cache",
            expire_after=timedelta(minutes=3),
        )

    def get_columns(self) -> dict[str, Field]:
        return {
            column.name: column.field
            for column in TABLES[self.base][self.resource]["columns"]
        }

    def get_data(
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        **kwargs: Any,
    ) -> Iterator[Row]:
        # apply default values
        for column in TABLES[self.base][self.resource]["columns"]:
            if column.default is not None and column.name not in bounds:
                bounds[column.name] = column.default

        if "number" in bounds:
            number = bounds.pop("number").value  # type: ignore
            return self._get_single_resource(number)

        return self._get_multiple_resources(bounds)

    def _get_single_resource(
        self,
        number: int,
    ) -> Iterator[Row]:
        """
        Return a specific resource.
        """
        headers = {"Accept": "application/vnd.github.v3+json"}
        if self.access_token:
            headers["Authorization"] = f"Bearer {self.access_token}"

        url = (
            f"https://api.github.com/{self.base}/{self.owner}/"
            f"{self.repo}/{self.resource}/{number}"
        )

        _logger.info("GET %s", url)
        response = self._session.get(url, headers=headers)
        payload = response.json()

        row = {
            column.name: get_value(column, payload)
            for column in TABLES[self.base][self.resource]["columns"]
        }
        row["rowid"] = 0
        _logger.debug(row)
        yield row

    def _get_multiple_resources(
        self,
        bounds: dict[str, Filter],
    ) -> Iterator[Row]:
        """
        Return multiple resources.
        """
        headers = {"Accept": "application/vnd.github.v3+json"}
        if self.access_token:
            headers["Authorization"] = f"Bearer {self.access_token}"

        url = f"https://api.github.com/{self.base}/{self.owner}/{self.repo}/{self.resource}"
        config = TABLES[self.base][self.resource]

        # map filters in ``bounds`` to query params
        params = {name: filter_.value for name, filter_ in bounds.items()}  # type: ignore

        page = 1
        rowid = 0
        while True:
            if config["paginated"]:
                _logger.info("GET %s (page %d)", url, page)
                params.update(
                    {
                        "per_page": PAGE_SIZE,
                        "page": page,
                    },
                )

            response = self._session.get(url, headers=headers, params=params)

            payload = response.json()
            if not payload:
                break

            if not response.ok:
                raise ProgrammingError(payload["message"])

            if processor := config["processor"]:
                payload = processor(payload)

            for resource in payload:
                row = {
                    column.name: get_value(column, resource)
                    for column in config["columns"]
                }
                row["rowid"] = rowid
                _logger.debug(row)
                yield row
                rowid += 1

            if not config["paginated"]:
                break

            page += 1


def get_value(column: Column, resource: dict[str, Any]) -> Any:
    """
    Extract the value of a column from a resource.
    """
    values = jsonpath.findall(column.json_path, resource)

    if isinstance(column.field, JSONString):
        return values

    try:
        return values[0]
    except IndexError:
        return None



"""
A simple example querying the Preset API.
"""

from shillelagh.backends.apsw.db import connect

if __name__ == "__main__":
    connection = connect(
        ":memory:",
        # create tokens/secrets at https://manage.app.preset.io/app/user
        adapter_kwargs={
            "presetapi": {
                "access_token": "",
                "access_secret": "",
            },
            "presetworkspaceapi": {
                "access_token": "",
                "access_secret": "",
            },
        },
    )
    cursor = connection.cursor()

    SQL = """
    SELECT * FROM
    "https://api.app.preset.io/v1/teams/"
    LIMIT 1
    """
    for row in cursor.execute(SQL):
        print(row)

    SQL = """
    SELECT * FROM
    "https://12345678.us1a.app.preset.io/api/v1/chart/"
    LIMIT 12
    """
    for row in cursor.execute(SQL):
        print(row)



"""
A simple example showing the Pandas adapter.
"""

import pandas as pd

from shillelagh.backends.apsw.db import connect

if __name__ == "__main__":
    connection = connect(":memory:")
    cursor = connection.cursor()

    mydf = pd.DataFrame({"a": [1, 2, 3]})

    SQL = "SELECT SUM(a) FROM mydf"
    for row in cursor.execute(SQL):
        print(row)

    SQL = "UPDATE mydf SET a = a + 1"
    cursor.execute(SQL)

    SQL = "SELECT SUM(a) FROM mydf"
    for row in cursor.execute(SQL):
        print(row)

    SQL = "SELECT * FROM mydf LIMIT 2 OFFSET 1"
    for row in cursor.execute(SQL):
        print(row)

    emptydf = pd.DataFrame({"a": []})
    SQL = "SELECT * from emptydf WHERE a = 'test'"
    for row in cursor.execute(SQL):
        print(row)



# pylint: disable=invalid-name, c-extension-no-member, unused-import
"""
A DB API 2.0 wrapper for APSW.
"""

import datetime
import logging
import re
from collections.abc import Iterator
from functools import partial
from typing import Any, Callable, Optional, TypeVar, cast

import apsw

from shillelagh import functions
from shillelagh.adapters.base import Adapter
from shillelagh.adapters.registry import registry
from shillelagh.backends.apsw.vt import VTModule, type_map
from shillelagh.db import (
    DEFAULT_SCHEMA,
    Connection,
    Cursor,
    apilevel,
    check_closed,
    paramstyle,
    threadsafety,
)
from shillelagh.exceptions import (  # nopycln: import; pylint: disable=redefined-builtin
    DatabaseError,
    DataError,
    Error,
    IntegrityError,
    InterfaceError,
    InternalError,
    NotSupportedError,
    OperationalError,
    ProgrammingError,
    Warning,
)
from shillelagh.fields import Blob, Field
from shillelagh.lib import (
    best_index_object_available,
    combine_args_kwargs,
    escape_identifier,
    find_adapter,
    serialize,
)
from shillelagh.types import (
    BINARY,
    DATETIME,
    NUMBER,
    ROWID,
    STRING,
    Binary,
    Date,
    DateFromTicks,
    Time,
    TimeFromTicks,
    Timestamp,
    TimestampFromTicks,
)
from shillelagh.typing import ColumnDescription, Description, SQLiteValidType

__all__ = [
    "DatabaseError",
    "DataError",
    "Error",
    "IntegrityError",
    "InterfaceError",
    "InternalError",
    "OperationalError",
    "NotSupportedError",
    "BINARY",
    "DATETIME",
    "NUMBER",
    "ROWID",
    "STRING",
    "Binary",
    "Date",
    "DateFromTicks",
    "Time",
    "TimeFromTicks",
    "Timestamp",
    "TimestampFromTicks",
    "Warning",
    "apilevel",
    "threadsafety",
    "paramstyle",
]

sqlite_version_info = tuple(
    int(number) for number in apsw.sqlitelibversion().split(".")
)

NO_SUCH_TABLE = re.compile("no such table: (?P<uri>.*)")

CURSOR_METHOD = TypeVar("CURSOR_METHOD", bound=Callable[..., Any])

_logger = logging.getLogger(__name__)


def get_missing_table(message: str) -> Optional[str]:
    """
    Return the missing table from a message.

    This is used to extract the table name from an APSW error message.
    """
    if match := NO_SUCH_TABLE.search(message):
        return match.groupdict()["uri"]

    return None


def get_type_code(type_name: str) -> type[Field]:
    """
    Return a ``Field`` that corresponds to a type name.

    This is used to build the description of the cursor after a successful
    query.
    """
    return cast(type[Field], type_map.get(type_name, Blob))


def convert_binding(binding: Any) -> SQLiteValidType:
    """
    Convert a binding to a SQLite type.

    Eg, if the user is filtering a timestamp column we need to convert the
    ``datetime.datetime`` object in the binding to a string.
    """
    if isinstance(binding, bool):
        return int(binding)
    if isinstance(binding, (int, float, str, bytes, type(None))):
        return binding
    if isinstance(binding, (datetime.datetime, datetime.date, datetime.time)):
        return binding.isoformat()
    return str(binding)


class APSWCursor(Cursor):  # pylint: disable=too-many-instance-attributes
    """
    Connection cursor.
    """

    def __init__(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        cursor: "apsw.Cursor",
        adapters: list[type[Adapter]],
        adapter_kwargs: dict[str, dict[str, Any]],
        isolation_level: Optional[str] = None,
        schema: str = DEFAULT_SCHEMA,
    ):
        super().__init__(adapters, adapter_kwargs, schema)

        self._cursor = cursor
        self.in_transaction = False
        self.isolation_level = isolation_level

        # Approach from: https://github.com/rogerbinns/apsw/issues/160#issuecomment-33927297
        # pylint: disable=unused-argument
        def exectrace(
            cursor: "apsw.Cursor",
            sql: str,
            bindings: Optional[tuple[Any, ...]],
        ) -> bool:
            # In the case of an empty sequence, fall back to None,
            # meaning no rows returned.
            self.description = self._cursor.getdescription() or None
            return True

        self._cursor.setexectrace(exectrace)

    @check_closed
    def execute(
        self,
        operation: str,
        parameters: Optional[tuple[Any, ...]] = None,
    ) -> "APSWCursor":
        """
        Execute a query using the cursor.
        """
        if not self.in_transaction and self.isolation_level:
            self._cursor.execute(f"BEGIN {self.isolation_level}")
            self.in_transaction = True

        self.description = None
        self._rowcount = -1

        # store current SQL in the cursor
        self.operation = operation

        # convert parameters (bindings) to types accepted by SQLite
        if parameters:
            parameters = tuple(convert_binding(parameter) for parameter in parameters)

        # this is where the magic happens: instead of forcing users to register
        # their virtual tables explicitly, we do it for them when they first try
        # to access them and it fails because the table doesn't exist yet
        created_tables = set()
        while True:
            try:
                self._cursor.execute(operation, parameters)
                self.description = self._get_description()
                self._results = self._convert(self._cursor)
                break
            except apsw.SQLError as ex:
                message = ex.args[0]
                if uri := get_missing_table(message):
                    if uri in created_tables:
                        raise ProgrammingError(message) from ex

                    # create the virtual table
                    self._create_table(uri)
                    created_tables.add(uri)
                    continue

                raise ProgrammingError(message) from ex

        if uri := self._drop_table_uri(operation):
            adapter, args, kwargs = find_adapter(
                uri,
                self._adapter_kwargs,
                self._adapters,
            )
            instance = adapter(*args, **kwargs)
            instance.drop_table()

        return self

    def _drop_table_uri(self, operation: str) -> Optional[str]:
        """
        Build a ``DROP TABLE`` regexp.
        """
        # remove comments
        operation = "\n".join(
            line for line in operation.split("\n") if not line.strip().startswith("--")
        )
        schema = re.escape(self.schema)
        regexp = re.compile(
            r"^\s*DROP\s+TABLE\s+(IF\s+EXISTS\s+)?"
            rf'({schema}\.)?(?P<uri>(.*?)|(".*?"))\s*;?\s*$',
            re.IGNORECASE,
        )
        if match := regexp.match(operation):
            return match.groupdict()["uri"].strip('"')

        return None

    def _convert(self, cursor: "apsw.Cursor") -> Iterator[tuple[Any, ...]]:
        """
        Convert row from SQLite types to native Python types.

        SQLite only supports 5 types. For booleans and time-related types
        we need to do the conversion here.
        """
        if not self.description:
            return  # pragma: no cover

        for row in cursor:
            yield tuple(
                # convert from SQLite types to native Python types
                type_map[desc[1].type]().parse(col)
                for col, desc in zip(row, self.description)
            )

    def _create_table(self, uri: str) -> None:
        """
        Create a virtual table.

        This method is called the first time a virtual table is accessed.
        """
        prefix = self.schema + "."
        if uri.startswith(prefix):
            uri = uri[len(prefix) :]

        adapter, args, kwargs = find_adapter(uri, self._adapter_kwargs, self._adapters)
        formatted_args = ", ".join(
            f"'{serialize(arg)}'"
            for arg in combine_args_kwargs(adapter, *args, **kwargs)
        )
        table_name = escape_identifier(uri)
        self._cursor.execute(
            f'CREATE VIRTUAL TABLE "{table_name}" USING {adapter.__name__}({formatted_args})',
        )

    def _get_description(self) -> Description:
        """
        Return the cursor description.

        We only return name and type, since that's what we get from APSW.
        """
        try:
            description = self._cursor.getdescription()
        except apsw.ExecutionCompleteError:
            return self.description

        annotated_alias = re.compile(r"^(\w+)\s*\[(\w+)]$")

        out: list[ColumnDescription] = []
        for name, type_name in description:
            if match := annotated_alias.match(name):
                name = match.group(1)
                type_name = match.group(2)

            type_code = get_type_code(type_name)
            out.append((name, type_code, None, None, None, None, True))

        return out

    @check_closed
    def close(self) -> None:
        """
        Close the cursor.

        This will also close the underlying APSW cursor.
        """
        if self.in_transaction:
            self._cursor.execute("ROLLBACK")
            self.in_transaction = False

        self._cursor.close()
        super().close()


def apsw_version() -> str:
    """
    Custom implementation of the ``VERSION`` function.

    This function shows the backend version::

        sql> SELECT VERSION();
        VERSION()
        ----------------------
        1.0.5 (apsw 3.36.0-r1)

    """
    return f"{functions.version()} (apsw {apsw.apswversion()})"


class APSWConnection(
    Connection[APSWCursor],
):  # pylint: disable=too-many-instance-attributes
    """Connection."""

    def __init__(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        path: str,
        adapters: list[type[Adapter]],
        adapter_kwargs: dict[str, dict[str, Any]],
        isolation_level: Optional[str] = None,
        apsw_connection_kwargs: Optional[dict[str, Any]] = None,
        schema: str = DEFAULT_SCHEMA,
        safe: bool = False,
    ):
        super().__init__(adapters, adapter_kwargs, schema, safe)

        # create underlying APSW connection
        apsw_connection_kwargs = apsw_connection_kwargs or {}
        self._connection = apsw.Connection(path, **apsw_connection_kwargs)
        self.isolation_level = isolation_level

        # register adapters
        for adapter in self._adapters:
            if best_index_object_available():
                self._connection.createmodule(
                    adapter.__name__,
                    VTModule(adapter),
                    use_bestindex_object=adapter.supports_requested_columns,
                )
            else:
                self._connection.createmodule(adapter.__name__, VTModule(adapter))

        # register functions
        available_functions = {
            "sleep": functions.sleep,
            "version": apsw_version,
            "get_metadata": partial(
                functions.get_metadata,
                self._adapter_kwargs,
                adapters,
            ),
            "date_trunc": functions.date_trunc,
        }
        if not safe:
            available_functions["upgrade"] = functions.upgrade

        for name, function in available_functions.items():
            self._connection.create_scalar_function(name, function)

    @check_closed
    def commit(self) -> None:
        """Commit any pending transaction to the database."""
        for cursor in self.cursors:
            if cursor.in_transaction:
                cursor._cursor.execute("COMMIT")  # pylint: disable=protected-access
                cursor.in_transaction = False

    @check_closed
    def rollback(self) -> None:
        """Rollback any transactions."""
        for cursor in self.cursors:
            if cursor.in_transaction:
                cursor._cursor.execute("ROLLBACK")  # pylint: disable=protected-access
                cursor.in_transaction = False

    @check_closed
    def cursor(self) -> APSWCursor:
        """Return a new Cursor Object using the connection."""
        cursor = APSWCursor(
            self._connection.cursor(),
            self._adapters,
            self._adapter_kwargs,
            self.isolation_level,
            self.schema,
        )
        self.cursors.append(cursor)

        return cursor


def connect(  # pylint: disable=too-many-arguments, too-many-positional-arguments
    path: str,
    adapters: Optional[list[str]] = None,
    adapter_kwargs: Optional[dict[str, dict[str, Any]]] = None,
    safe: bool = False,
    isolation_level: Optional[str] = None,
    apsw_connection_kwargs: Optional[dict[str, Any]] = None,
    schema: str = DEFAULT_SCHEMA,
) -> APSWConnection:
    """
    Constructor for creating a connection to the database.
    """
    adapter_kwargs = adapter_kwargs or {}
    enabled_adapters = registry.load_all(adapters, safe)

    # replace entry point names with class names
    mapping = {
        name: adapter.__name__.lower() for name, adapter in enabled_adapters.items()
    }
    adapter_kwargs = {mapping[k]: v for k, v in adapter_kwargs.items() if k in mapping}

    return APSWConnection(
        path,
        list(enabled_adapters.values()),
        adapter_kwargs,
        isolation_level,
        apsw_connection_kwargs,
        schema,
        safe,
    )



"""
SQLAlchemy dialect.
"""

# pylint: disable=abstract-method, unused-argument

from types import ModuleType
from typing import Any, Optional, TypedDict, cast

import sqlalchemy.types
from sqlalchemy.engine.base import Connection as SqlaConnection
from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy.pool.base import _ConnectionFairy
from sqlalchemy.sql import compiler
from sqlalchemy.sql.type_api import TypeEngine

from shillelagh.adapters.base import Adapter
from shillelagh.backends.sqlglot import db
from shillelagh.backends.sqlglot.db import Connection
from shillelagh.exceptions import ProgrammingError
from shillelagh.lib import find_adapter


class SQLAlchemyColumn(TypedDict):
    """
    A custom type for a SQLAlchemy column.
    """

    name: str
    type: TypeEngine
    nullable: bool
    default: Optional[str]
    autoincrement: str
    primary_key: int


class SQLGlotDialect(DefaultDialect):
    """
    A SQLAlchemy dialect for Shillelagh based on the Python executor.
    """

    name = "shillelagh"
    driver = "sqlglot"

    statement_compiler = compiler.SQLCompiler
    ddl_compiler = compiler.DDLCompiler
    type_compiler = compiler.GenericTypeCompiler
    preparer = compiler.IdentifierPreparer

    supports_alter = False
    supports_comments = True
    inline_comments = True
    supports_statement_cache = True

    supports_schemas = False
    supports_views = False
    postfetch_lastrowid = False

    supports_native_boolean = True

    isolation_level = "AUTOCOMMIT"

    default_paramstyle = "qmark"

    supports_is_distinct_from = False

    @classmethod
    def dbapi(cls) -> ModuleType:  # pylint: disable=method-hidden
        """
        Return the DB API module.
        """
        return db

    @classmethod
    def import_dbapi(cls):  # pylint: disable=method-hidden
        """
        Return the DB API module.
        """
        return db

    def __init__(
        self,
        adapters: Optional[list[str]] = None,
        adapter_kwargs: Optional[dict[str, dict[str, Any]]] = None,
        safe: bool = False,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs or {}
        self._safe = safe

    def do_ping(self, dbapi_connection: _ConnectionFairy) -> bool:
        """
        Return true if the database is online.
        """
        return True

    def has_table(  # pylint: disable=unused-argument
        self,
        connection: _ConnectionFairy,
        table_name: str,
        schema: Optional[str] = None,
        **kwargs: Any,
    ) -> bool:
        """
        Return true if a given table exists.
        """
        try:
            get_adapter_for_table_name(connection, table_name)
            return True
        except ProgrammingError:
            pass

        return False

    def get_table_names(
        self,
        connection: SqlaConnection,
        schema: Optional[str] = None,
        **kw: Any,
    ) -> list[str]:
        """
        Return a list of table names.
        """
        return []

    def get_columns(  # pylint: disable=unused-argument
        self,
        connection: _ConnectionFairy,
        table_name: str,
        schema: Optional[str] = None,
        **kwargs: Any,
    ) -> list[SQLAlchemyColumn]:
        adapter = get_adapter_for_table_name(connection, table_name)
        columns = adapter.get_columns()
        return [
            {
                "name": column_name,
                "type": getattr(sqlalchemy.types, field.type),
                "nullable": True,
                "default": None,
                "autoincrement": "auto",
                "primary_key": 0,
            }
            for column_name, field in columns.items()
        ]

    def do_rollback(self, dbapi_connection: Connection) -> None:
        """
        Executor doesn't support rollbacks.
        """

    # methods that are needed for integration with Apache Superset
    def get_schema_names(self, connection: SqlaConnection, **kw: Any):
        """
        Return the list of schemas.
        """
        return ["main"]

    def get_pk_constraint(
        self,
        connection: SqlaConnection,
        table_name: str,
        schema: Optional[str] = None,
        **kw: Any,
    ):
        return {"constrained_columns": [], "name": None}

    def get_foreign_keys(
        self,
        connection: SqlaConnection,
        table_name: str,
        schema: Optional[str] = None,
        **kw: Any,
    ):
        return []

    get_check_constraints = get_foreign_keys
    get_indexes = get_foreign_keys
    get_unique_constraints = get_foreign_keys

    def get_table_comment(self, connection, table_name, schema=None, **kwargs):
        return {"text": ""}


def get_adapter_for_table_name(
    connection: _ConnectionFairy,
    table_name: str,
) -> Adapter:
    """
    Return an adapter associated with a connection.

    This function instantiates the adapter responsible for a given table name,
    using the connection to properly pass any adapter kwargs.
    """
    raw_connection = cast(db.Connection, connection.engine.raw_connection())
    adapter, args, kwargs = find_adapter(
        table_name,
        raw_connection._adapter_kwargs,  # pylint: disable=protected-access
        raw_connection._adapters,  # pylint: disable=protected-access
    )
    return adapter(*args, **kwargs)  # type: ignore
